#include "CGame.h"

#define STB_IMAGE_IMPLEMENTATION
#include "stb_image/stb_image.h"
#include "Model.h"
#include "Mesh.h"
#include "glm/mat4x4.hpp"
#include "glm/vec3.hpp"
#include "glm/glm.hpp"
#include "glm/gtc/matrix_transform.hpp"
#include "glm/gtc/type_ptr.hpp"
#include <chrono>

void CGame::Initialize(unsigned int screenWidth, unsigned int screenHeight)
{
    m_Width = screenWidth;
    m_Height = screenHeight;

    SetupGPUProgram();
    SetupModel();

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_STENCIL_TEST);
}

void CGame::Shutdown()
{
    
}

void CGame::OnResize(int width, int height)
{
    m_Width = width;
    m_Height = height;

    glViewport(0, 0, m_Width, m_Height);
    glScissor(0, 0, m_Width, m_Height);
}

void CGame::Render()
{
    static auto startTime = std::chrono::high_resolution_clock::now();

    auto currentTime = std::chrono::high_resolution_clock::now();
    float time = std::chrono::duration<float, std::chrono::seconds::period>(currentTime - startTime).count();

    glClearColor(0.2f, 0.3f, 0.3f, 1.0f);
    glClearDepth(1.0f);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glColorMask(1, 1, 1, 1);

    // Draw the Filled Man
    glEnable(GL_STENCIL_TEST);
    glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
    glStencilFunc(GL_ALWAYS, 0, 0xFF); // all fragments should update the stencil buffer
    glStencilMask(0xFF); // enable writing to the stencil buffer
    glm::mat4 model = glm::rotate(glm::mat4(1.0f), time * glm::radians(90.0f), glm::vec3(0.0f, 1.0f, 0.0f));
    glm::mat4 view;
    view = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, -2.0f, -10.0f));
    glm::mat4 projection;
    projection = glm::perspective(glm::radians(60.0f), m_Width / static_cast<float>(m_Height), 0.03f, 100.0f);
    glm::mat4 mvp = projection * view * model;

    m_pShaderPass0->use();
    m_pShaderPass0->setMatrix4x4("modelMatrix", glm::value_ptr(model));
    m_pShaderPass0->setMatrix4x4("viewMatrix", glm::value_ptr(view));
    m_pShaderPass0->setMatrix4x4("projectionMatrix", glm::value_ptr(projection));
    m_pShaderPass0->setMatrix4x4("mvpMatrix", glm::value_ptr(mvp));
    m_pShaderPass0->setFloat("time", time);
    glDepthFunc(GL_LEQUAL);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    m_pManModel->meshes[0].Draw(*m_pShaderPass0);

    // Draw the Stencil Quad
    glStencilMask(0xFF);
    model = glm::rotate(glm::mat4(1.0f), glm::radians(90.0f), glm::vec3(1.0f, 0.0f, 0.0f));
    model = glm::translate(model, glm::vec3(0.0f, 4.0f, -2.5f));
    model = glm::scale(model, glm::vec3(1.0f, 1.0f, 1.0f));
    mvp = projection * view * model;
    m_pShaderStencilPass->use();
    m_pShaderStencilPass->setFloat("time", time);
    m_pShaderStencilPass->setMatrix4x4("mvpMatrix", glm::value_ptr(mvp));
    glDepthFunc(GL_ALWAYS);
    glDepthMask(0);
    glStencilFunc(GL_ALWAYS, 1, 0xFF);
    glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
    glColorMask(0, 0, 0, 0);
    m_pStencilModel->meshes[0].Draw(*m_pShaderStencilPass);
    glDepthFunc(GL_LEQUAL);
    glDepthMask(1);
    glColorMask(1, 1, 1, 1);

    // Draw the Wireframe man
    model = glm::rotate(glm::mat4(1.0f), time * glm::radians(90.0f), glm::vec3(0.0f, 1.0f, 0.0f));
    view = glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, -2.0f, -10.0f));
    projection = glm::perspective(glm::radians(60.0f), m_Width / static_cast<float>(m_Height), 0.03f, 100.0f);
    mvp = projection * view * model;
    m_pShaderPass1->use();
    m_pShaderPass1->setMatrix4x4("modelMatrix", glm::value_ptr(model));
    m_pShaderPass1->setMatrix4x4("viewMatrix", glm::value_ptr(view));
    m_pShaderPass1->setMatrix4x4("projectionMatrix", glm::value_ptr(projection));
    m_pShaderPass1->setMatrix4x4("mvpMatrix", glm::value_ptr(mvp));
    m_pShaderPass1->setFloat("time", time);

    glStencilFunc(GL_EQUAL, 1, 0xFF);
    glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);

    glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    m_pManModel->meshes[0].Draw(*m_pShaderPass1);
}

void CGame::SetupModel()
{
    m_pManModel = new Model();
    m_pManModel->Load("man.obj");

    m_pStencilModel = new Model();
    m_pStencilModel->Load("quad.obj");
}

void CGame::SetupGPUProgram()
{
    m_pShaderPass0 = new Shader("vertexPass0.glsl", "fragmentPass0.glsl");
    m_pShaderPass1 = new Shader("vertexPass1.glsl", "fragmentPass1.glsl");
    m_pShaderStencilPass = new Shader("vertexStencil.glsl", "fragmentStencil.glsl");
}